<?
require_once("dbh.php");

define("DB_PATH", "db");
define("LDBI_INDEX", "1");
define("LDBI_TABLE", "2");
define("LDBI_SEQUENCE", "3");

class LDBI
{
	function LDBI($name = "database", $create = false)
	{
		$path = sprintf("%s/%s", DB_PATH, $name);
		if($create)
		{
			if(!file_exists($path))
				$ret = mkdir($path);
		}
		else $ret = file_exists($path);

		$this->path = $path;

		return $ret;
	}

	function load_tables()
	{
		/* this function open all tables in read/write mode */
	}

	function &table($name, $mode = "c")
	{
		$t = &new LDBI_table(&$this, $name, $this->path."/".$name.".db", $mode);
		$t->type = LDBI_TABLE;

		$this->tables[$name] =& $t;

		return $t;
	}

	function &sequence($name, $mode = "c")
	{
		$seq =& new LDBI_sequence($this, $name, $mode);

		$seq->type = LDBI_SEQUENCE;

		return $seq;
	}
}

class LDBI_table
{
	function LDBI_table(&$db, $name = "table", $path = "table.db", $mode = "c")
	{
		$this->dbh =& $db;
		$this->path = $path;
		$this->name = $name;
		$this->mode = $mode;
		$this->lock($path);
		if(!($this->open($path, $mode)))
		{
			return false;
		}
	}

	function destroy()
	{
		if($this->type != LDBI_INDEX)
		{
			$this->update();
		}
		$this->unlock();
		if($this->mode != "r")
			dbh_optimize($this->db);
		$this->close();
	}

	function lock($path)
	{
		if(!file_exists($path)) return;

		$fp = fopen($path.".rlck", "w");
		if($fp)
		{
			$tries = 1;
			$timedout = FALSE;
			$finished = FALSE;
			while(!$timedout && !$finished)
			{
				if($tries >= 7)
				{
					$timedout = TRUE;
				}
				$lk = flock($fp, LOCK_EX, FALSE);
				if($lk)
				{
					$this->lock_fp = $fp;
					$finished = TRUE;
				}
				else
				{
					sleep($tries);
				}
			}
		}
	}

	function unlock()
	{
		flock($this->lock_fp, LOCK_UN, FALSE);
		fclose($this->lock_fp);
		unlink($this->path.".rlck");
	}

	function open($path, $mode)
	{
		$this->db = dbh_open($path, $mode);
		if($this->mode != "r")
			dbh_optimize($this->db);
		return $this->db;
	}

	function close()
	{
		dbh_close($this->db);
	}

	function insert($value, $key = false)
	{
		$insv['d'] = $value;
		$insv['m'] = 1;
		if(!$key) $key = $this->getnewkey();
		if(!dbh_exists($this->db, $key))
			dbh_insert($this->db, $key, serialize($insv));
		else echo "ALREADY EXIST!\n";
	}

	function cleaninsert($value, $key = false)
	{
		if(!$key) $key = $this->getnewkey();
		if(!dbh_exists($this->db, $key))
			dbh_insert($this->db, $key, serialize($value));
		else echo "ALREADY EXIST!\n";
	}

	function &buildindex()
	{
		global $my__index;
		$multiple = false;
		$cols = false;
		$cols = func_get_args();
		if(is_array($cols[0])) $cols = $cols[0];
		$name = join("-", $cols);
		$db =& $this->dbh->table($this->name.".idx_on_$name", "c");
		$db->type = LDBI_INDEX;
		if($db)
			$info = $this->load();

		if($info)
		{
			foreach($info as $key=>$val)
			{
				if(!$val['m']) continue;
				$this->keyring[] = $key;
				$val = $val['d'];
				$index = array();
				for($i = 0 ; $i < count($cols) ; $i++)
				{
					$col = $cols[$i];
					$ival = $val[$col];
					$index[$col][] = $ival;

					if(!is_string($ival))
					{
						continue;
					}

					$index[$col][] = "%".$ival."%";

					for($p = 1 ; $p < strlen($ival) ; $p++)
					{
						$ivalp = false;
						$ivalp = substr($ival, $p);
						$index[$col][] ="%".$ivalp."%";
					}

					for($p = 2 ; $p < strlen($ival) ; $p++)
					{
						$ivalp = false;
						$ivalp = substr($ival, -$p);
						$index[$col][] ="%".$ivalp."%";
					}

					for($p = 1 ; $p < ((strlen($ival) / 2) - 1) ; $p++, $l++)
					{
						$ivalp = false;
						$ivalp = substr($ival, $p, strlen($ival) - ($p * 2));
						$index[$col][]= "%".$ivalp."%";
					}

					for($p = 0 ; $p < strlen($ival) ; $p++, $l++)
					{
						$ivalp = false;
						$ivalp = substr($ival, $p, 1);
						$index[$col][]= "%".$ivalp."%";
					}
				}

				idx_sort($index);
				$my__index = array();
				$mycols = array();
				foreach($index as $myk=>$myv)
				{
					$mycols[] = $myk;
				}
				$mycols[] = 0;
				reset($index);
				makeindex($index, 0, $mycols);
				reset($index);
				reset($my__index);
				foreach($my__index as $rowv)
					$db->idx_insert(serialize($rowv),$key, $this);
			}
		}
		$db->destroy();
	}

	function &rebuildindex()
	{
		$args = func_get_args();
		$this->update(1);
		return $this->buildindex($args);
	}

	function idx_insert($key, $value, $table)
	{
		$values = $this->idx_search($key);

		for($i = 0 ; $i < count($values) ; $i++)
		{
			if($values[$i] == $value) return;
		}

		if($values) $alreadye = true;
		$values[] = $value;

		if($alreadye)
			dbh_replace($this->db, $key, serialize($values));
		else
			$this->cleaninsert($values, $key);
	}

	function idx_search($key)
	{
		return $this->get($key);
	}

	function replace($value, $where)
	{
		$keys = $this->search($where);

		if($key)
		{
			foreach($keys as $i=>$row)
			{
				$repv = $this->get($row);
				$repv['m'] = 1;
				foreach($value as $k=>$v)
				{
					$repv['d'][$k] = $v;
				}
				reset($value);
				dbh_replace($this->db, $row, serialize($repv));
			}
		}

	}

	function update($flag = 0)
	{
		if(is_array($this->keyring))
		foreach($this->keyring as $key)
		{
			$v = $this->get($key);
			if($v)
			{
				$v['m'] = $flag;
				dbh_replace($this->db, $key, serialize($v));
			}
		}
		else
		{
			$k = dbh_firstkey($this->db);
			while($k)
			{
				$v = $this->get($k);
				if($v)
				{
					$v['m'] = $flag;
					dbh_replace($this->db, $key,
						serialize($v));
				}
				$k = dbh_nextkey($this->db);
			}
		}
	}

	function delete($where)
	{
		$keys = $this->search($where);
		if($keys)
		{
			foreach($keys as $key)
				dbh_delete($this->db, $key);
		}
	}

	function parse_where_clause($where)
	{
		if(!preg_match("/and/i", $where))
		{
			return array($where);
		}

		return preg_split("/and/i", $where, PREG_SPLIT_NO_EMPTY);
	}

	function search($where)
	{
		$wcs = $this->parse_where_clause($where);
		foreach($wcs as $wc)
		{
			list($col, $val) = split("=", trim($wc));
			$col = trim($col);
			$val = trim($val);
			$colsa[] = $col;
			$s[$col] = $val;
		}

		$cols = join("-", $colsa);

		$idx =& $this->dbh->table($this->name.".idx_on_$cols", "r");
		$idx->type = LDBI_INDEX;
		if($idx->db)
		{
			$ret = $idx->get(serialize($s));
			$idx->destroy();
		}
		else
		{
			$tmp = $this->load();
			foreach($tmp as $key=>$value)
			{
				if($value['d'][$col] == $val)
				{
					$ret[] = $key;
					break;
				}
			}
		}

		return $ret;
	}

	function load($path = false)
	{
		if(!$this->db && $path) open($path, "c");

		if($this->db)
		{
			$key = dbh_firstkey($this->db);
			while($key)
			{
				$ret[$key] = unserialize(dbh_fetch($this->db, $key));
				$key = dbh_nextkey($this->db);
			}
		}

		return $ret;
	}

	function get($key)
	{
		return unserialize(dbh_fetch($this->db, $key));
	}

	function getnewkey()
	{
		$tmpk = dbh_firstkey($this->db);
		while($tmpk)
		{
			if($tmpk > $key) $key = $tmpk;
			$tmpk = dbh_nextkey($this->db);
		}

		$key++;

		return $key;
	}
}

class LDBI_sequence
{
	function LDBI_sequence(&$db, $name = "table", $mode = "c")
	{
		$this->name = $name;
		$this->db =& $db;
		$this->table =& $db->table($name."_seq", $mode);
	}

	function next()
	{
		$seqno = $this->table->get(1);
		$seqno = $seqno['d']['seqno'];
		if(!$seqno) $seqno = 0;
		$seqno++;
		$r['m'] = 0;
		$r['d'] = array("seqno"=>$seqno);
		dbh_replace($this->table->db, 1, serialize($r));

		return $seqno;
	}
}

function makeindex(&$ind, $depth = 0, $cols, $prev = false)
{
	global $my__index;
	$tmp = false;

	$key = $cols[$depth];

	if($key)
	{
		$val = $ind[$key];
		for($i = 0 ; $i < count($val) ; $i++)
		{
			$prev[$key] = $val[$i];
			makeindex(&$ind, $depth + 1, $cols, $prev);
		}
	}
	else if(is_array($prev))
	{
		$my__index[] = $prev;
	}
}

function idx_sort($ind)
{
	usort($ind, "idx_cmp");
}

function idx_cmp($a, $b)
{
	return count($a) - count($b);
}

?>
